"""
    Copyright (C) 2022 - Rokas Puzonas <rokas.puz@gmail.com>

    This file is part of Class Diagram Generator.

    KTU OOP Report Generator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    KTU OOP Report Generator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KTU OOP Report Generator. If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Iterable, Iterator, Optional
from lark.lark import Lark
from lark.lexer import Token
from lark.tree import ParseTree, Tree
from .class_diagram import EnumDiagram, NamespaceDiagram, VisibilityEnum, ParameterDirection, ClassAttribute, ClassMethodParameter, ClassMethod, ClassDiagram

# NOTE: This parser assumes that the c# file has not syntax errors.
# Because I am not validating if something looks logical or not.

# BUG: Methods only support strings and numbers as default parameter values

l = Lark(r"""
    start: namespace_import* namespace_definition

    IDENTIFIER_NAME: "@"? /[a-zA-Z0-9_\.]+/
    !type_name: IDENTIFIER_NAME
             | IDENTIFIER_NAME "[" ","* "]"
             | IDENTIFIER_NAME "<" (type_name ",")* type_name ">"

    SIGNED_INTEGER: /[+-]?(0|[1-9][0-9]*)/

    ANY_VALUE: /.+/

    namespace_import: "using" IDENTIFIER_NAME ";"

    namespace_definition: "namespace" IDENTIFIER_NAME "{" namespace_block "}"
    namespace_block: (class_definition | enum_definition | interface_definition)*

    VISIBILITY: "private" | "public" | "protected"
    modifier: (VISIBILITY | "static" | "abstract")~1..2

    enum_definition.2: modifier? "enum" IDENTIFIER_NAME [":" IDENTIFIER_NAME] "{" enum_block? "}"
    enum_block: (IDENTIFIER_NAME ",")* IDENTIFIER_NAME

    class_definition.2: (modifier | "partial" | "sealed" | "internal" | "abstract")* "class" type_name [":" class_inhereted_classes] [class_constraints] "{" class_block "}"
    class_inhereted_classes: (type_name ",")* type_name
    class_constraints: "where" IDENTIFIER_NAME ":" (type_name ",")* type_name
    class_block: (method | attribute | class_definition)*

    interface_definition.2: modifier* "interface" type_name "{" _method_block* "}"

    attribute_method_modifier: (modifier | "const" | "readonly" | "override" | "virtual")~1..5

    attribute.2: attribute_method_modifier? type_name IDENTIFIER_NAME (("=" ATTRIBUTE_VALUE) | ";")
               | attribute_method_modifier? type_name IDENTIFIER_NAME "{" attribute_getset "}" ["=" ATTRIBUTE_VALUE]
    ATTRIBUTE_VALUE: /.+/ ";"
    attribute_getset: attribute_getter? attribute_setter?
                    | attribute_setter? attribute_getter?
    attribute_getter: VISIBILITY? "get" ("{" _method_block* "}")? ";"?
    attribute_setter: VISIBILITY? "set" ("{" _method_block* "}")? ";"?

    method.2: attribute_method_modifier? return_type? IDENTIFIER_NAME [method_constraints] "(" method_parameters? ")" method_where_constraints* method_body
            | attribute_method_modifier? return_type? IDENTIFIER_NAME [method_constraints] "(" method_parameters? ")" method_constructor_call? method_body
            | attribute_method_modifier? return_type? "operator" OPERATOR "(" method_parameters? ")" method_body
    !return_type: "void" | type_name
    method_body: "{" _method_block* "}"
               | ";"
    method_constraints.2: "<" (IDENTIFIER_NAME ",")* IDENTIFIER_NAME ">"
    method_where_constraints.2: "where" IDENTIFIER_NAME ":" (type_name ",")* type_name
    method_constructor_call.2: ":" IDENTIFIER_NAME "(" (IDENTIFIER_NAME ",")* IDENTIFIER_NAME ")"
    method_parameters.2: (method_parameter ",")* method_parameter
    method_parameter: METHOD_DIRECTION? "params"? type_name IDENTIFIER_NAME ["=" METHOD_PARAMETER_DEFAULT]
    METHOD_DIRECTION: "ref" | "out"
    METHOD_PARAMETER_DEFAULT: SIGNED_INTEGER | STRING | "null"
    _method_block: /[^{}]+/ | ("{" _method_block* "}")
    OPERATOR: "<" | ">" | "==" | "!=" | ">=" | "<="

    DECORATOR: "[" IDENTIFIER_NAME "]"

    %import common.WS
    %import common.C_COMMENT
    %import common.CPP_COMMENT
    %import common.ESCAPED_STRING -> STRING
    %ignore C_COMMENT
    %ignore CPP_COMMENT
    %ignore DECORATOR
    %ignore WS
""")

def find_direct_child_token(tree: ParseTree, token: str) -> Optional[Token]:
    for child in tree.children:
        if isinstance(child, Token) and child.type == token:
            return child

def find_direct_child_rule(tree: ParseTree, token: str) -> Optional[ParseTree]:
    for child in tree.children:
        if isinstance(child, Tree) and child.data == token:
            return child

def find_direct_data(tree: ParseTree, data: str) -> Iterator[ParseTree]:
    for child in tree.children:
        if isinstance(child, Tree) and child.data == data:
            yield child
        elif isinstance(child, Token) and child.type == data:
            yield child # type: ignore

def find_namespace_names(tree: ParseTree) -> Iterable[tuple[ParseTree, str]]:
    for namespace_node in tree.find_data("namespace_definition"):
        name = find_direct_child_token(namespace_node, "IDENTIFIER_NAME")
        assert name
        yield (namespace_node, name.value)

def find_class_names(tree: ParseTree) -> Iterable[tuple[ParseTree, str]]:
    for class_node in find_direct_data(tree, "class_definition"):
        # print(find_type_name(class_node))
        # name = find_direct_child_token(class_node, "IDENTIFIER_NAME")
        name = find_type_name(class_node)
        assert name
        yield (class_node, name)

        class_block = find_direct_child_rule(class_node, "class_block")
        assert class_block
        for (sub_class_node, sub_class_name) in find_class_names(class_block):
            yield (sub_class_node, name+"."+sub_class_name)

def find_visibility(tree: ParseTree) -> Optional[VisibilityEnum]:
    modifier = find_direct_child_rule(tree, "modifier")
    if not modifier:
        return VisibilityEnum.Private

    visibility = find_direct_child_token(modifier, "VISIBILITY")
    if not visibility:
        return VisibilityEnum.Private

    if visibility.value == "public":
        return VisibilityEnum.Public
    elif visibility.value == "private":
        return VisibilityEnum.Private
    elif visibility.value == "protected":
        return VisibilityEnum.Protected
    else:
        return VisibilityEnum.Private

def find_type_name(tree: ParseTree, token: str = "type_name") -> Optional[str]:
    type_name = find_direct_child_rule(tree, token)
    if not type_name:
        return None

    return "".join(type_name.scan_values(lambda v: isinstance(v, Token)))

def find_attributes(tree: ParseTree) -> Iterable[ClassAttribute]:
    class_block = find_direct_child_rule(tree, "class_block")
    assert class_block
    for attr_node in find_direct_data(class_block, "attribute"):
        name = find_direct_child_token(attr_node, "IDENTIFIER_NAME")
        assert name
        attr_type = find_type_name(attr_node)
        assert attr_type
        visibility = find_visibility(attr_node)
        assert visibility
        initial_value = find_direct_child_token(attr_node, "ATTRIBUTE_VALUE")
        if initial_value:
            initial_value = initial_value.value[:-1].strip()
        yield ClassAttribute(name.value, attr_type, visibility, initial_value)

def find_method_parameters(tree: ParseTree) -> list[ClassMethodParameter]:
    parameters_node = find_direct_child_rule(tree, "method_parameters")
    if not parameters_node:
        return []

    parameters = []
    for parameter_node in parameters_node.find_data("method_parameter"):
        parameter_name = find_direct_child_token(parameter_node, "IDENTIFIER_NAME")
        assert parameter_name

        parameter_type = find_type_name(parameter_node, "type_name")
        assert parameter_type

        parameter_default = find_direct_child_token(parameter_node, "METHOD_PARAMETER_DEFAULT")
        if parameter_default:
            parameter_default = parameter_default.value

        direction = ParameterDirection.In
        direction_node = find_direct_child_token(parameter_node, "METHOD_DIRECTION")
        if direction_node:
            if direction_node.value == "ref":
                direction = ParameterDirection.InOut
            elif direction_node.value == "out":
                direction = ParameterDirection.Out

        parameters.append(ClassMethodParameter(
            name = parameter_name.value,
            type = parameter_type,
            direction = direction,
            default = parameter_default
        ))

    return parameters

def find_methods(tree: ParseTree) -> Iterable[ClassMethod]:
    class_block = find_direct_child_rule(tree, "class_block")
    assert class_block
    for method_node in find_direct_data(class_block, "method"):
        name = find_direct_child_token(method_node, "IDENTIFIER_NAME")

        # If it doesn't have a name, it means that this is an overloaded operator
        if name:
            name = name.value
            method_constraints = find_direct_child_rule(method_node, "method_constraints")
            if method_constraints:
                name += "<"
                name += ",".join(method_constraints.scan_values(lambda v: isinstance(v, Token)))
                name += ">"

            return_type = find_type_name(method_node, "return_type")

            visibility = find_visibility(method_node)
            assert visibility

            parameters = find_method_parameters(method_node)

            yield ClassMethod(name, return_type, visibility, parameters)

def find_enum_names(tree: ParseTree) -> Iterable[tuple[ParseTree, str]]:
    for enum_node in tree.find_data("enum_definition"):
        name = find_direct_child_token(enum_node, "IDENTIFIER_NAME")
        assert name
        yield (enum_node, name.value)

def extract_enum_values(tree: ParseTree) -> list[str]:
    enum_block = find_direct_child_rule(tree, "enum_block")
    assert enum_block

    values = []
    for value_node in enum_block.children:
        assert isinstance(value_node, Token)
        values.append(value_node.value)

    return values

def extract_namespaces(filename: str) -> list[NamespaceDiagram]:
    contents = None
    with open(filename, "r", encoding="utf-8-sig") as f:
        contents = f.read()
    assert contents

    namespaces = []
    tree = l.parse(contents)
    for (namespace_node, namespace_name) in find_namespace_names(tree):
        namespace_block = find_direct_child_rule(namespace_node, "namespace_block")
        assert namespace_block

        classess = []
        for (class_node, class_name) in find_class_names(namespace_block):
            classess.append(ClassDiagram(
                name = class_name,
                attributes = list(find_attributes(class_node)),
                methods = list(find_methods(class_node))
            ))

        enums = []
        for (enum_node, enum_name) in find_enum_names(namespace_block):
            enums.append(EnumDiagram(
                name = enum_name,
                values = extract_enum_values(enum_node)
            ))

        namespaces.append(NamespaceDiagram(
            name = namespace_name,
            classess = classess,
            enums = enums
        ))
    return namespaces



"""
    Copyright (C) 2022 - Rokas Puzonas <rokas.puz@gmail.com>

    This file is part of Class Diagram Generator.

    KTU OOP Report Generator is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    KTU OOP Report Generator is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with KTU OOP Report Generator. If not, see <https://www.gnu.org/licenses/>.
"""
from enum import Enum
from dataclasses import dataclass
import math
from PIL import Image, ImageFont, ImageDraw
from math import floor
from typing import Literal, Optional

TARGET_RATIO = 1.8

class VisibilityEnum(Enum):
    Public = 1,
    Private = 2,
    Protected = 3

class ParameterDirection(Enum):
    In = 1,
    Out = 2,
    InOut = 3

@dataclass
class ClassAttribute:
    name: str
    type: str
    visibility: VisibilityEnum
    initial_value: Optional[str]

@dataclass
class ClassMethodParameter:
    name: str
    type: str
    direction: ParameterDirection
    default: Optional[str]

@dataclass
class ClassMethod:
    name: str
    return_type: Optional[str]
    visibility: VisibilityEnum
    parameters: list[ClassMethodParameter]

@dataclass
class ClassDiagram:
    name: str
    attributes: list[ClassAttribute]
    methods: list[ClassMethod]

    def __str__(self) -> str:
        return f"ClassDiagram({self.name})"

@dataclass
class EnumDiagram:
    name: str
    values: list[str]

    def __str__(self) -> str:
        return f"EnumDiagram({self.name})"

@dataclass
class NamespaceDiagram:
    name: str
    classess: list[ClassDiagram]
    enums: list[EnumDiagram]

def merge_similar_classess(diagrams: list[ClassDiagram]):
    grouped_diagrams = {}

    for diagram in diagrams:
        if diagram.name not in grouped_diagrams:
            grouped_diagrams[diagram.name] = []
        grouped_diagrams[diagram.name].append(diagram)

    for key in grouped_diagrams.keys():
        if len(grouped_diagrams[key]) > 1:
            group: list[ClassDiagram] = grouped_diagrams[key]
            first_diagram = group[0]
            for other_diagram in group[1:]:
                first_diagram.methods.extend(other_diagram.methods)
                first_diagram.attributes.extend(other_diagram.attributes)
                diagrams.remove(other_diagram)

def merge_similar_namespaces(namespaces: list[NamespaceDiagram]):
    grouped_namespaces = {}

    for diagram in namespaces:
        if diagram.name not in grouped_namespaces:
            grouped_namespaces[diagram.name] = []
        grouped_namespaces[diagram.name].append(diagram)

    for key in grouped_namespaces.keys():
        if len(grouped_namespaces[key]) > 1:
            group: list[NamespaceDiagram] = grouped_namespaces[key]
            first_diagram = group[0]
            for other_diagram in group[1:]:
                first_diagram.classess.extend(other_diagram.classess)
                first_diagram.enums.extend(other_diagram.enums)
                namespaces.remove(other_diagram)

    for diagram in namespaces:
        merge_similar_classess(diagram.classess)

def stringify_visibility(visibility: VisibilityEnum) -> str:
    if visibility == VisibilityEnum.Protected:
        return "#"
    elif visibility == VisibilityEnum.Public:
        return "+"
    elif visibility == VisibilityEnum.Private:
        return "-"

def stringify_class_attributes(attributes: list[ClassAttribute]) -> list[str]:
    attribute_lines = []

    for attr in attributes:
        attr_line = ""

        attr_line += stringify_visibility(attr.visibility)
        attr_line += " " + attr.name
        attr_line += ": " + attr.type
        if attr.initial_value:
            attr_line += " = " + attr.initial_value

        attribute_lines.append(attr_line)

    return attribute_lines

def stringify_class_parameters(parameters: list[ClassMethodParameter]) -> str:
    stringified_parameters = []
    for param in parameters:
        param_str = ""
        if param.direction == ParameterDirection.In:
            param_str += "in"
        elif param.direction == ParameterDirection.Out:
            param_str += "out"
        elif param.direction == ParameterDirection.InOut:
            param_str += "inout"

        param_str += " " + param.name
        param_str += ": " + param.type
        if param.default:
            param_str += " = " + param.default
        stringified_parameters.append(param_str)
    return ", ".join(stringified_parameters)

def stringify_class_methods(methods: list[ClassMethod]) -> list[str]:
    method_lines = []

    for method in methods:
        method_line = ""

        method_line += stringify_visibility(method.visibility)
        parameters = stringify_class_parameters(method.parameters)
        method_line += f" {method.name}({parameters})"
        if method.return_type and method.return_type != "void":
            method_line += ": " + method.return_type

        method_lines.append(method_line)

    return method_lines

def render_diagram_sections(
        sections: list[list[str]],
        font_file: str,
        font_size: int,
        background: str,
        foreground: str,
        padding: int,
        border_width: int,
        border_color: str,
        line_spacing: float
    ) -> Image.Image:

    font = ImageFont.truetype(font_file, font_size)

    max_line_width = 0
    for section in sections:
        if len(section) > 0:
            max_line_width = max(max_line_width, max(font.getlength(s) for s in section))

    total_line_amount = sum(max(len(section), 1) for section in sections)

    line_height = font_size * line_spacing
    total_width = 2*border_width + 2*padding + floor(max_line_width)
    total_height = 2*border_width + \
                   floor(total_line_amount*line_height - len(sections)*(font_size*(line_spacing-1))) + \
                   len(sections)*(border_width+2*padding)

    # Create image and start drawing
    image = Image.new("RGB", (total_width, total_height), background)

    draw = ImageDraw.Draw(image)
    # Create a cursor, so it is easier to track where things need to be drawn next
    cx = border_width + padding # Cursor x coordinate
    cy = border_width + padding # Cursor y coordinate

    # Draw borders
    draw.rectangle((0, 0, total_width, total_height), outline=border_color, width=border_width)

    # For each section...
    for i in range(len(sections)):
        section = sections[i]

        # Draw all of it's entries line by line and...
        for line in range(len(section)):
            pos = (cx, cy+line*line_height)
            draw.text(pos, section[line], fill=foreground, font=font)

        cy += padding + max(len(section), 1)*line_height + border_width/2 - (font_size*(line_spacing-1))

        # IF this isin't the last section, draw a seperating line
        if i != len(sections)-1:
            draw.line([(0, cy), (total_width-1, cy)], border_color, border_width)

        cy += padding + border_width/2

    return image

def render_class_diagram(
        diagram: ClassDiagram,
        font_file: str,
        font_size: int,
        background: str = "#FFFFFF",
        foreground: str = "#000000",
        padding: int = 10,
        border_width: int = 5,
        border_color: str = "#000000",
        line_spacing: float = 1.20
    ) -> Image.Image:
    sections = []

    sections.append([diagram.name])
    sections.append(stringify_class_attributes(diagram.attributes))
    sections.append(stringify_class_methods(diagram.methods))

    return render_diagram_sections(
        sections,
        font_file, font_size, background, foreground, padding, border_width,
        border_color, line_spacing
    )

def render_enum_diagram(
        diagram: EnumDiagram,
        font_file: str,
        font_size: int,
        background: str = "#FFFFFF",
        foreground: str = "#000000",
        padding: int = 10,
        border_width: int = 5,
        border_color: str = "#000000",
        line_spacing: float = 1.20
    ) -> Image.Image:
    sections = []

    sections.append([diagram.name])
    values_section = []
    for value in diagram.values:
        values_section.append(f"+ {value}")
    sections.append(values_section)

    return render_diagram_sections(
        sections,
        font_file, font_size, background, foreground, padding, border_width,
        border_color, line_spacing
    )

def place_images_into_square(
        images: list[Image.Image],
        spacing: int,
        sort_key,
        cut: Literal["vertical"]|Literal["horizontal"] = "vertical",
        reverse_cut: bool = False
    ) -> tuple[list[tuple[float, float, Image.Image]], float, float]:
    images.sort(key=sort_key, reverse=True)

    available_areas = [(0.0, 0.0, math.inf, math.inf)]
    image_positions: list[tuple[float, float, Image.Image]] = []
    packed_width = 0.0
    packed_height = 0.0

    for image in images:
        best_area_index = -1
        best_area_ratio = 0

        # Find which available area is most sutable for image to be placed in,
        # so that the packed images would be in a square
        for i in range(len(available_areas)):
            area = available_areas[i]
            area_width = area[2] - area[0]
            area_height = area[3] - area[1]
            if image.width <= area_width and image.height <= area_height:
                possible_packed_width = max(area[0] + image.width, packed_width)
                possible_packed_height = max(area[1] + image.height, packed_height)
                possible_ratio = possible_packed_width / possible_packed_height
                if best_area_index == -1 or abs(TARGET_RATIO-possible_ratio) < abs(TARGET_RATIO-best_area_ratio):
                    best_area_ratio = possible_ratio
                    best_area_index = i

        assert best_area_index > -1

        # When a sutable area has been found, save a position where the image
        # would be placed, and cut up the area so the next image placement
        # don't overlap already placed image
        left, top, right, bottom = available_areas.pop(best_area_index)
        image_positions.append((left, top, image))

        packed_width = max(left + image.width, packed_width)
        packed_height = max(top + image.height, packed_height)

        new_area_1 = None
        new_area_2 = None
        if cut == "vertical":
            new_area_1 = (left+image.width+spacing, top, right, bottom)
            new_area_2 = (left, top+image.height+spacing, left + image.width, bottom)
        else:
            new_area_1 = (left, top+image.height + spacing, right, bottom)
            new_area_2 = (left + image.width + spacing, top, right, top + image.height)

        if reverse_cut:
            available_areas.append(new_area_1)
            available_areas.append(new_area_2)
        else:
            available_areas.append(new_area_2)
            available_areas.append(new_area_1)

    return (image_positions, packed_width, packed_height)

def find_best_image_placements_in_square(images: list[Image.Image], spacing: int) -> tuple[list[tuple[float, float, Image.Image]], float, float]:
    possible_sort_keys = [
        lambda image: image.height*image.width,
        lambda image: image.height,
        lambda image: image.width
    ]

    best_image_positions = []
    best_width = 0.0
    best_height = 1.0

    for sort_key in possible_sort_keys:
        for cut in ("vertical", "horizontal"):
            for reverse_cut in (False, True):
                image_positions, width, height = place_images_into_square(
                    images,
                    spacing,
                    sort_key,
                    cut, # type: ignore
                    reverse_cut
                )

                ratio = width/height
                best_ratio = best_width/best_height

                if best_image_positions == None or abs(TARGET_RATIO-ratio) < abs(TARGET_RATIO-best_ratio):
                    best_image_positions = image_positions
                    best_width = width
                    best_height = height

    return (best_image_positions, best_width, best_height)

# Because the area into which there images are being packed is unbounded,
# I picked a criteria so that it tries to create a packed image which is as
# close a square as possible.
def pack_images(images: list[Image.Image], background: str, spacing: int=0) -> Image.Image:
    image_positions, width, height = find_best_image_placements_in_square(images, spacing)

    packed_image = Image.new("RGB", (int(width), int(height)), background)

    for x, y, image in image_positions:
        packed_image.paste(image, (int(x), int(y)))

    return packed_image

def render_namespaces(
        namespaces: list[NamespaceDiagram],
        font_file: str,
        font_size: int,
        background: str = "#FFFFFF",
        foreground: str = "#000000",
        diagram_padding: int = 10,
        border_width: int = 5,
        border_color: str = "#000000",
        line_spacing: float = 1.20,
        diagram_spacing: int = 25
    ) -> Image.Image:
    rendered_diagrams: list[Image.Image] = []

    for namespace in namespaces:
        for diagram in namespace.classess:
            rendered_diagrams.append(render_class_diagram(
                diagram,
                font_file, font_size, background, foreground, diagram_padding, border_width,
                border_color, line_spacing
            ))

        for diagram in namespace.enums:
            rendered_diagrams.append(render_enum_diagram(
                diagram,
                font_file, font_size, background, foreground, diagram_padding, border_width,
                border_color, line_spacing
            ))

    return pack_images(rendered_diagrams, background, diagram_spacing)



import setuptools

with open("README.md", "r") as f:
	long_description = f.read()

setuptools.setup(
	name="class-diagram-generator",
	version="2.0.10",
	author="Rokas Puzonas",
	author_email="rokas.puz@gmail.com",
	description="Generate standardized class diagrams based on C# source code",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/RokasPuzonas/class-diagram-generator",
	project_urls={
		"Source Code": "https://github.com/RokasPuzonas/class-diagram-generator",
	},
	license="GNU GPLv2",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"Topic :: Utilities",
		"Intended Audience :: Developers",
	],
	install_requires=["lark", "Pillow"],
	python_requires=">=3.10",
)



from os.path import join, relpath, isfile, isdir
from os import walk
from fnmatch import fnmatch
from typing import Iterable, Optional
from PIL import Image

from .class_diagram import render_namespaces, merge_similar_namespaces
from .csharp_analyser import extract_namespaces

def is_file_included(filename: str, included: list[str], excluded: list[str]) -> bool:
    if not any(fnmatch(filename, pattern) for pattern in included):
        return False

    if any(fnmatch(filename, pattern) for pattern in excluded):
        return False

    return True

def list_files(folder_path: str, included: list[str], excluded: list[str]) -> Iterable[str]:
    for (root, _, files) in walk(folder_path, topdown=True):
        for name in files:
            full_path = join(root, name)
            if is_file_included(relpath(full_path, folder_path), included, excluded):
                yield full_path

def render_from_paths(*paths: str, font_file: str, font_size: int) -> Optional[Image.Image]:
    included_files = ["*.cs"]
    excluded_files = ["*.designer.cs", "**/obj/**", "obj/**", "**/bin/**", "bin/**", "Properties/**", "**/Properties/**", "**/AssemblyInfo.cs"]
    namespaces = []

    for path in paths:
        if isfile(path):
            if is_file_included(path, included_files, excluded_files):
                for diagram in extract_namespaces(path):
                    namespaces.append(diagram)
        elif isdir(path):
            for filename in list_files(path, included_files, excluded_files):
                for diagram in extract_namespaces(filename):
                    namespaces.append(diagram)


    merge_similar_namespaces(namespaces)

    if len(namespaces) > 0:
        return render_namespaces(namespaces, font_file, font_size)
    else:
        return None
